"""
Amatino API Python Bindings
Tree Test Module
Author: hugh@blinkybeach.com
"""
from amatino.tests.primary.account import AccountTest
from amatino import Account
from amatino import Transaction
from amatino import Tree
from amatino import AMType
from amatino import Side
from amatino import Entry
from amatino import GlobalUnit
from amatino import TreeNode
from amatino import Entity
from datetime import datetime
from datetime import timedelta
from decimal import Decimal

NAME = 'Retrieve a Tree'


class TreeTest(AccountTest):
    """Test the Performance object"""

    def __init__(self, name=NAME) -> None:
        super().__init__(name)
        return

    def execute(self) -> None:

        try:

            income = self.create_account(
                amt=AMType.income,
                name='Test Income'
            )

            income_child = Account.create(
                self.entity,
                'Test income child',
                income.am_type,
                income.denomination,
                parent=income
            )

            expense = self.create_account(
                amt=AMType.expense,
                name='Test Expense'
            )

            asset = self.create_account(
                AMType.asset,
                'Test Asset'
            )

            Transaction.create(
                self.entity,
                datetime.utcnow(),
                [
                    Entry(Side.credit, Decimal('400'), income),
                    Entry(Side.debit, Decimal('400'), asset)
                ],
                self.usd,
                'Test transaction 1'
            )

            Transaction.create(
                self.entity,
                datetime.utcnow(),
                [
                    Entry(Side.credit, Decimal('250'), income_child),
                    Entry(Side.credit, Decimal('250'), income),
                    Entry(Side.debit, Decimal('500'), asset)
                ],
                self.usd,
                'Test transaction 2'
            )

            Transaction.create(
                self.entity,
                datetime.utcnow(),
                [
                    Entry(Side.credit, Decimal('300'), asset),
                    Entry(Side.debit, Decimal('300'), expense)
                ],
                self.usd,
                'Test transaction 3'
            )

            Transaction.create(
                self.entity,
                datetime.utcnow() + timedelta(days=4),
                [
                    Entry(Side.debit, Decimal('50'), expense),
                    Entry(Side.credit, Decimal('50'), asset)
                ],
                self.usd,
                'Test transaction 4'
            )

            tree = Tree.retrieve(
                entity=self.entity,
                balance_time=datetime.utcnow() + timedelta(hours=1),
                denomination=self.usd
            )

            assert isinstance(tree, Tree)
            assert tree.has_income
            assert tree.has_expenses
            assert not tree.has_liabilities
            assert not tree.has_equity
            assert tree.global_unit_id == self.usd.id_
            assert tree.custom_unit_id is None
            assert isinstance(tree.denomination, GlobalUnit)
            assert isinstance(tree.balance_time, datetime)
            assert isinstance(tree.generated_time, datetime)
            assert isinstance(tree.entity, Entity)

            income_root = tree.income[0]
            expense_root = tree.expenses[0]

            assert isinstance(income_root, TreeNode)
            assert isinstance(expense_root, TreeNode)
            assert income_root.has_children
            assert not expense_root.has_children
            assert isinstance(income_root.children[0], TreeNode)

            assert income_root.recursive_balance == Decimal('900')
            assert income_root.account_balance == Decimal('650')
            assert income_root.children[0].account_balance == Decimal('250')
            assert expense_root.account_balance == Decimal('300')
            assert expense_root.recursive_balance == Decimal('300')

            assert tree.total_income == Decimal('900')
            assert tree.total_expenses == Decimal('300')
            assert tree.total_assets == Decimal('600')
            assert tree.total_equity == Decimal('0')
            assert tree.total_liabilities == Decimal('0')

        except Exception as error:
            self.record_failure(error)
            return

        self.record_success()



"""
Amatino Python
Session Decodable Module
author: hugh@blinkybeach.com
"""
from typing import Any, Optional, TypeVar, Type, List
from amatino.session import Session

T = TypeVar('T', bound='SessionDecodable')


class SessionDecodable:
    """Abstract protocol defining an interface for decodable classes"""

    @classmethod
    def decode(self, data: Any, session: Session) -> T:
        """Return a JSON-serialisable form of the object"""
        raise NotImplementedError

    @classmethod
    def optionally_decode(
        cls: Type[T],
        data: Optional[Any],
        session: Session
    ) -> Optional[T]:
        """Optionally return a decoded object from serialised data"""
        if data is None:
            return None
        return cls.decode(data, session)

    @classmethod
    def decode_many(cls: Type[T], data: Any, session: Session) -> List[T]:
        """Return list of decoded instances of an object"""
        return [cls.decode(d, session) for d in data]

    @classmethod
    def optionally_decode_many(
        cls: Type[T],
        data: Optional[Any],
        session: Session,
        default_to_empty_list: bool = False
    ) -> Optional[List[T]]:
        """Optionally return a list of decoded objects"""
        if data is None and default_to_empty_list is True:
            return list()
        if data is None:
            return None
        return cls.decode_many(data, session)



"""
Amatino API Python Bindings
User Module
Author: hugh@amatino.io
"""
from amatino.internal.immutable import Immutable
from amatino.internal.am_time import AmatinoTime
from amatino import User, Session, State
from typing import List, Type, TypeVar, Any, Optional
from amatino.internal.api_request import ApiRequest
from amatino.internal.url_parameters import UrlParameters
from amatino.internal.url_target import UrlTarget
from amatino.internal.http_method import HTTPMethod
from amatino.api_error import ApiError
from amatino.missing_key import MissingKey
from collections.abc import Sequence

T = TypeVar('T', bound='UserList')


class UserList(Sequence):
    """
    A User List is a collection of Users for whom the retrieving User has
    billing responsibility, and who were created using the Amatino API.

    The User List excludes Users managed by the billing dashboard.
    """
    _PATH = '/users/list'

    def __init__(
        self,
        page: int,
        number_of_pages: int,
        generated_time: AmatinoTime,
        state: State,
        users: List[User],
        session: Session
    ) -> None:

        assert isinstance(generated_time, AmatinoTime)
        assert isinstance(number_of_pages, int)
        assert isinstance(page, int)
        assert isinstance(users, list)
        assert False not in [isinstance(u, User) for u in users]
        assert isinstance(state, State)
        assert isinstance(session, Session)

        self._generated_time = generated_time
        self._number_of_pages = number_of_pages
        self._page = page
        self._users = users
        self._state = state
        self._session = session

        return

    generated_time = Immutable(lambda s: s._generated_time.raw)
    number_of_pages = Immutable(lambda s: s._number_of_pages)
    page = Immutable(lambda s: s._page)
    users = Immutable(lambda s: s._users)
    state = Immutable(lambda s: s._state)
    has_more_pages = Immutable(lambda s: s._number_of_pages > self._page)

    def __iter__(self):
        return UserList.Iterator(self._users)

    class Iterator:
        """An iterator for iterating through users in a UserList"""

        def __init__(self, users: List[User]) -> None:
            self._index = 0
            self._users = users
            return

        def __next__(self) -> User:
            if self._index >= len(self._users):
                raise StopIteration
            user = self._users[self._index]
            self._index += 1
            return user

    def __len__(self):
        return len(self.users)

    def __getitem__(self, key):
        return self.users[key]

    def next_page(self: T) -> Optional[T]:
        """
        Return the next page available in this Entity List, or None if no more
        pages are available.
        """
        if not self.has_more_pages:
            return None

        return self.retrieve(
            session=self._session,
            state=self._state,
            page=self._page + 1
        )

    @classmethod
    def retrieve(
        cls: Type[T],
        session: Session,
        state: State = State.ALL,
        page: int = 1
    ) -> T:
        """
        Retrieve a UserList from the perspective of the User tied to the
        supplied Session. Optionally specify a State (active, all, deleted) and
        integer page number.
        """
        if not isinstance(session, Session):
            raise TypeError('session must be of type Session')

        if not isinstance(state, State):
            raise TypeError('state must be of type State')

        if not isinstance(page, int):
            raise TypeError('page must be of type int')

        state_target = UrlTarget('state', state.value)
        page_target = UrlTarget('page', str(page))
        parameters = UrlParameters(targets=[state_target, page_target])

        request = ApiRequest(
            path=cls._PATH,
            method=HTTPMethod.GET,
            credentials=session,
            data=None,
            url_parameters=parameters
        )

        return cls.decode(session, request.response_data)

    @classmethod
    def decode(
        cls: Type[T],
        session: Session,
        data: Any
    ) -> T:

        if not isinstance(data, dict):
            raise ApiError('Unexpected data type returned: ' + str(type(data)))

        try:
            user_list = cls(
                page=data['page'],
                number_of_pages=data['number_of_pages'],
                generated_time=AmatinoTime.decode(data['generated_time']),
                state=State(data['state']),
                users=User.decode_many(session, data['users']),
                session=session
            )
        except KeyError as error:
            raise MissingKey(error.args[0])

        return user_list



"""
Amatino API Python Bindings
API Request Module
Author: hugh@amatino.io

This module is intended to be private, used indirectly
by public classes, and should not be used directly.
"""
import sys
from json import loads
from urllib.request import Request
from urllib.request import urlopen
from urllib.error import HTTPError
from amatino.internal.credentials import Credentials
from amatino.internal.data_package import DataPackage
from amatino.internal.url_parameters import UrlParameters
from amatino.internal.request_headers import RequestHeaders
from amatino.internal.http_method import HTTPMethod
from typing import Optional
from amatino.internal.immutable import Immutable
from amatino.internal.errors.not_found import ResourceNotFound


class ApiRequest:
    """
    Private - Not intended to be used directly.

    An instance of an http request to the Amatino API.
    """

    _ENDPOINT = 'https://api.amatino.io'
    _DEBUG_ENDPOINT = 'http://127.0.0.1:5000'
    _TIMEOUT = 10

    def __init__(
        self,
        path: str,
        method: HTTPMethod,
        credentials: Optional[Credentials] = None,
        data: Optional[DataPackage] = None,
        url_parameters: Optional[UrlParameters] = None,
        debug: bool = False
    ) -> None:

        self._response_data = None

        if credentials is not None:
            assert isinstance(credentials, Credentials)

        if data is not None:
            assert isinstance(data, DataPackage)
            request_data = data.as_json_bytes()
        else:
            request_data = None

        if url_parameters is not None:
            assert isinstance(url_parameters, UrlParameters)

        if debug is True or '--amatino-debug' in sys.argv[1:]:
            url = self._DEBUG_ENDPOINT
        else:
            url = self._ENDPOINT

        url += path

        if url_parameters is not None:
            url += url_parameters.parameter_string()

        headers = RequestHeaders(path, credentials, data)
        request = Request(
            url=url,
            data=request_data,
            headers=headers.dictionary(),
            method=method.value,
        )
        try:
            self._response = urlopen(request, timeout=self._TIMEOUT)
        except HTTPError as error:
            if error.code == 404:
                raise ResourceNotFound
            raise error

        self._response_data = loads(self._response.read().decode('utf-8'))

        return

    response_data = Immutable(lambda s: s._response_data)



"""
Amatino API Python Bindings
Signature Module
Author: hugh@amatino.io

This module is intended to be private, used indirectly by public classes, and
should not be used directly.
"""
from hashlib import sha512
from base64 import b64encode
import hmac
import time


class Signature:
    """An HMAC signature computation"""

    def __init__(
        self,
        api_key: str,
        path: str,
    ) -> None:

        assert isinstance(api_key, str)
        assert isinstance(path, str)

        encoded_key = api_key.encode('utf-8')

        timestamp = str(int(time.time()))

        message = timestamp + path

        encoded_message = message.encode('utf-8')

        digest = hmac.new(encoded_key, encoded_message, sha512).digest()
        self._signature = b64encode(digest).decode()

        return

    string = property(lambda s: s._signature)



"""
Amatino API Python Bindings
Transaction Version List Test Module
Author: hugh@blinkybeach.com
"""
from amatino.tests.primary.account import AccountTest
from amatino import TransactionVersionList
from amatino import Account
from amatino import Transaction
from amatino import AMType
from amatino import Side
from amatino import Entry
from datetime import datetime
from datetime import timedelta
from decimal import Decimal

NAME = 'Retrieve a TransactionVersionList'


class TxVersionListTest(AccountTest):
    """Test the Performance object"""

    def __init__(self, name=NAME) -> None:
        super().__init__(name)
        return

    def execute(self) -> None:

        try:

            income = self.create_account(
                amt=AMType.income,
                name='Test Income'
            )

            income_child = Account.create(
                self.entity,
                'Test income child',
                income.am_type,
                income.denomination,
                parent=income
            )

            expense = self.create_account(
                amt=AMType.expense,
                name='Test Expense'
            )

            transaction = Transaction.create(
                self.entity,
                datetime.utcnow(),
                [
                    Entry(Side.credit, Decimal('400'), income),
                    Entry(Side.debit, Decimal('400'), expense)
                ],
                self.usd,
                'Test transaction 1'
            )

            transaction.update(
                time=datetime.utcnow(),
                entries=[
                    Entry(Side.credit, Decimal('250'), income_child),
                    Entry(Side.credit, Decimal('250'), income),
                    Entry(Side.debit, Decimal('500'), expense)
                ],
                denomination=self.usd,
                description='Test transaction v2'
            )

            transaction.update(
                time=datetime.utcnow() - timedelta(days=4),
                entries=[
                    Entry(Side.credit, Decimal('1'), income),
                    Entry(Side.debit, Decimal('1'), expense)
                ],
                denomination=self.usd,
                description='Test transaction v3'
            )

            tx_version_list = TransactionVersionList.retrieve(
                entity=self.entity,
                transaction=transaction
            )

            assert len(tx_version_list) == 3, 'list length 3'
            for version in tx_version_list:
                assert isinstance(version, Transaction), 'full of Transactions'

        except Exception as error:
            self.record_failure(error)
            return

        self.record_success()



"""
Amatino API Python Bindings
User List Test Module
Author: hugh@amatino.io
"""
from amatino.tests.ancillary.session import SessionTest
from amatino import UserList
from amatino import User
from amatino import State


class UserListTest(SessionTest):
    """Test the retrieval of a UserList"""
    _NAME = 'Retrieve a UserList'

    def __init__(self) -> None:
        super().__init__(self._NAME)
        return

    def execute(self) -> None:

        self.create_session()

        try:
            user_list = UserList.retrieve(
                self.session,
                state=State.ALL
            )
            assert isinstance(user_list, UserList)
            for user in user_list:
                assert isinstance(user, User)
        except Exception as error:
            self.record_failure(error)
            return

        self.record_success()
        return



"""
Amatino API Python Bindings
Ledger Module
Author: hugh@amatino.io
"""
from datetime import datetime
from amatino.ledger_order import LedgerOrder
from amatino.internal.am_time import AmatinoTime
from amatino.denomination import Denomination
from amatino.custom_unit import CustomUnit
from amatino.global_unit import GlobalUnit
from amatino.session import Session
from amatino.entity import Entity
from amatino.internal.immutable import Immutable
from amatino.account import Account
from amatino.internal.encodable import Encodable
from amatino.internal.api_request import ApiRequest
from amatino.internal.url_parameters import UrlParameters
from amatino.ledger_row import LedgerRow
from amatino.unexpected_response_type import UnexpectedResponseType
from amatino.missing_key import MissingKey
from amatino.internal.http_method import HTTPMethod
from amatino.internal.data_package import DataPackage
from amatino.internal.am_amount import AmatinoAmount
from typing import Optional
from typing import TypeVar
from typing import Type
from typing import Dict
from typing import Any
from typing import List
from collections import Sequence
from amatino.denominated import Denominated

T = TypeVar('T', bound='Ledger')


class Ledger(Sequence, Denominated):
    """
    A Ledger is a list of Transactions from the perspective of a particular
    Account. Ledgers are ordered by Transaction time, and include a running
    Account Balance for every line.

    You can request Ledgers in arbitrary Global or Custom Units, not just the
    native unit of the target Account. If you request a Ledger in a unit other
    than the target Account native unit, Amatino will compute and return
    unrealised gains and losses.

    Amatino will return a maximum total of 1,000 Ledger Rows per retrieval
    request. If the Ledger you define spans more than 1,000 rows, it will be
    broken into pages you can retrieve seperately.
    """

    _PATH = '/accounts/ledger'

    def __init__(
        self,
        entity: Entity,
        account_id: int,
        start_time: AmatinoTime,
        end_time: AmatinoTime,
        recursive: bool,
        generated_time: AmatinoTime,
        global_unit_id: Optional[int],
        custom_unit_id: Optional[int],
        page: int,
        number_of_pages: int,
        order: LedgerOrder,
        ledger_rows: List[LedgerRow]
    ) -> None:

        self._entity = entity
        self._account_id = account_id
        self._start_time = start_time
        self._end_time = end_time
        self._recursive = recursive
        self._generated_time = generated_time
        self._global_unit_id = global_unit_id
        self._custom_unit_id = custom_unit_id
        self._page = page
        self._number_of_pages = number_of_pages
        self._order = order
        self._rows = ledger_rows

        return

    session = Immutable(lambda s: s._entity.session)
    entity = Immutable(lambda s: s._entity)
    account_id = Immutable(lambda s: s._account_id)
    account = Immutable(
        lambda s: Account.retrieve(s.session, s.entity, s.account_id)
    )
    start_time = Immutable(lambda s: s._start_time.raw)
    end_time = Immutable(lambda s: s._end_time.raw)
    recursive = Immutable(lambda s: s._recursive)
    generated_time = Immutable(lambda s: s._generated_time.raw)
    global_unit_id = Immutable(lambda s: s._global_unit_id)
    custom_unit_id = Immutable(lambda s: s._custom_unit_id)
    page = Immutable(lambda s: s._page)
    number_of_pages = Immutable(lambda s: s._number_of_pages)
    order = Immutable(lambda s: s._order)
    rows = Immutable(lambda s: s._rows)

    @classmethod
    def retrieve(
        cls: Type[T],
        entity: Entity,
        account: Account,
        order: LedgerOrder = LedgerOrder.YOUNGEST_FIRST,
        page: int = 1,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        denomination: Optional[Denomination] = None
    ) -> T:
        """
        Retrieve a Ledger for the supplied account. Optionally specify order,
        page, denomination, start time, and end time.
        """
        if not isinstance(entity, Entity):
            raise TypeError('entity must be of type `Entity`')

        arguments = Ledger.RetrieveArguments(
            account,
            order,
            page,
            start_time,
            end_time,
            denomination
        )
        data = DataPackage(object_data=arguments, override_listing=True)

        parameters = UrlParameters(entity_id=entity.id_)

        request = ApiRequest(
            path=cls._PATH,
            method=HTTPMethod.GET,
            credentials=entity.session,
            data=data,
            url_parameters=parameters
        )

        return cls._decode(entity, request.response_data)

    @classmethod
    def _decode(
        cls: Type[T],
        entity: Entity,
        data: Any
    ) -> T:

        if not isinstance(data, dict):
            raise UnexpectedResponseType(type(data), dict)

        try:
            ledger = cls(
                entity=entity,
                account_id=data['account_id'],
                start_time=AmatinoTime.decode(data['start_time']),
                end_time=AmatinoTime.decode(data['end_time']),
                recursive=data['recursive'],
                generated_time=AmatinoTime.decode(data['generated_time']),
                global_unit_id=data['global_unit_denomination'],
                custom_unit_id=data['custom_unit_denomination'],
                ledger_rows=Ledger._decode_rows(data['ledger_rows']),
                page=data['page'],
                number_of_pages=data['number_of_pages'],
                order=LedgerOrder(data['ordered_oldest_first'])
            )
        except KeyError as error:
            raise MissingKey(error.args[0])

        return ledger

    @classmethod
    def _decode_rows(cls: Type[T], rows: List[Any]) -> List[LedgerRow]:
        """Return LedgerRows decoded from raw API response data"""
        if not isinstance(rows, list):
            raise UnexpectedResponseType(rows, list)

        def decode(data) -> LedgerRow:

            if not isinstance(data, list):
                raise UnexpectedResponseType(data, list)

            row = LedgerRow(
                transaction_id=data[0],
                transaction_time=AmatinoTime.decode(data[1]),
                description=data[2],
                opposing_account_id=data[3],
                opposing_account_name=data[4],
                debit=AmatinoAmount.decode(data[5]),
                credit=AmatinoAmount.decode(data[6]),
                balance=AmatinoAmount.decode(data[7])
            )

            return row

        return [decode(r) for r in rows]

    class RetrieveArguments(Encodable):
        def __init__(
            self,
            account: Account,
            order: LedgerOrder = LedgerOrder.YOUNGEST_FIRST,
            page: int = 1,
            start_time: Optional[datetime] = None,
            end_time: Optional[datetime] = None,
            denomination: Optional[Denomination] = None
        ) -> None:

            if not isinstance(account, Account):
                raise TypeError('account must be of type `Account`')

            if not isinstance(order, LedgerOrder):
                raise TypeError('order must beof type `LedgerOrder`')

            if not isinstance(page, int):
                raise TypeError('page must be of type `int`')

            if start_time and not isinstance(start_time, datetime):
                raise TypeError('start_time must be of type `datetime` or None')

            if end_time and not isinstance(end_time, datetime):
                raise TypeError('end_time must be of type `datetime` or None')

            if denomination and not isinstance(denomination, Denomination):
                raise TypeError(
                    'denomination must be of type `Denomination` or None'
                )

            if denomination is None:
                denomination = account.denomination

            self._account = account
            self._order = order
            self._page = page
            self._start_time = None
            if start_time:
                self._start_time = AmatinoTime(start_time)
            self._end_time = None
            if end_time:
                self._end_time = AmatinoTime(end_time)
            self._denomination = denomination

        def serialise(self) -> Dict[str, Any]:
            global_unit_id = None
            custom_unit_id = None
            if isinstance(self._denomination, GlobalUnit):
                global_unit_id = self._denomination.id_
            else:
                assert isinstance(self._denomination, CustomUnit)
                custom_unit_id = self._denomination.id_

            start_time = None
            if self._start_time:
                start_time = AmatinoTime(self._start_time).serialise()

            end_time = None
            if self._end_time:
                end_time = AmatinoTime(self._end_time).serialise()

            data = {
                'account_id': self._account.id_,
                'start_time': start_time,
                'end_time': end_time,
                'page': self._page,
                'global_unit_denomination': global_unit_id,
                'custom_unit_denomination': custom_unit_id,
                'order_oldest_first': self._order.value
            }

            return data

    def __iter__(self):
        return Ledger.Iterator(self._rows)

    class Iterator:
        """An iterator for iterating through LedgerRows in a Ledger"""

        def __init__(self, rows: List[LedgerRow]) -> None:
            self._index = 0
            self._rows = rows
            return

        def __next__(self) -> LedgerRow:
            if self._index >= len(self._rows):
                raise StopIteration
            row = self._rows[self._index]
            self._index += 1
            return row

    def __len__(self):
        return len(self.rows)

    def __getitem__(self, key):
        return self.rows[key]



"""
Amatino API Python Bindings
HTTP Method Module
Author: hugh@amatino.io
"""
from enum import Enum


class HTTPMethod(Enum):
    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    PATCH = 'PATCH'
    DELETE = 'DELETE'



"""
Amatino API Python Bindings
User Test Module
Author: hugh@amatino.io
"""
from amatino import User, ResourceNotFound
from amatino.tests.ancillary.session import SessionTest


class UserTest(SessionTest):
    """Test the User ancillary object"""

    def __init__(self, name='Create, retrieve, delete Users') -> None:

        super().__init__(name)
        self.create_session()
        return

    def execute(self) -> None:

        try:
            user = User.retrieve(self.session, self.user_id)
            assert isinstance(user, User), 'retrieved is User'

            new_user = User.create(
                session=self.session,
                secret='what a great passphrase',
                handle='Dick Cheney'
            )
            assert isinstance(new_user, User), 'new user is User'

            new_user.delete()

            was_deleted = False
            try:
                User.retrieve(self.session, new_user.id_)
            except ResourceNotFound:
                was_deleted = True

        except Exception as error:
            self.record_failure(error)
            return

        if was_deleted is False:
            self.record_failure('New user not deleted')

        self.record_success()
        return



"""
Amatino API Python Bindings
Tree Module
Author: hugh@amatino.io
"""
from typing import Type
from typing import TypeVar
from typing import List
from typing import Optional
from typing import Any
from typing import Dict
from datetime import datetime
from decimal import Decimal
from amatino.am_type import AMType
from amatino.denominated import Denominated
from amatino.denomination import Denomination
from amatino.decodable import Decodable
from amatino.internal.data_package import DataPackage
from amatino.internal.api_request import ApiRequest
from amatino.internal.url_parameters import UrlParameters
from amatino.internal.http_method import HTTPMethod
from amatino.internal.encodable import Encodable
from amatino.missing_key import MissingKey
from amatino.entity import Entity
from amatino.unexpected_response_type import UnexpectedResponseType
from amatino.internal.am_time import AmatinoTime
from amatino.tree_node import TreeNode
from amatino.internal.immutable import Immutable
from amatino.global_unit import GlobalUnit
from amatino.custom_unit import CustomUnit

T = TypeVar('T', bound='Tree')
K = TypeVar('K', bound='Tree.RetrieveArguments')


class Tree(Decodable, Denominated):
    """
    Trees present the entire chart of Accounts of an Entity in a single
    hierarchical object.

    Each Account is nested under its parent, and in turn lists all its
    children, all providing Balances and Recursive Balances.

    Trees are trimmed for permissions. If the user requesting the tree only
    has read access to a subset of Accounts, they will only receive a tree
    containing those accounts, with placeholder objects filling the place
    of Accounts they are not permitted to read.

    Each Account in the Tree is presented as a Tree Node.
    """
    _PATH = '/trees'

    def __init__(
        self,
        entity: Entity,
        balance_time: AmatinoTime,
        generated_time: AmatinoTime,
        global_unit_denomination: Optional[int],
        custom_unit_denomination: Optional[int],
        tree: List[TreeNode]
    ) -> None:

        assert isinstance(entity, Entity)
        assert isinstance(balance_time, AmatinoTime)
        assert isinstance(generated_time, AmatinoTime)
        if global_unit_denomination is not None:
            assert isinstance(global_unit_denomination, int)
        if custom_unit_denomination is not None:
            assert isinstance(custom_unit_denomination, int)
        assert isinstance(tree, list)
        assert False not in [isinstance(t, TreeNode) for t in tree]

        self._entity = entity
        self._balance_time = balance_time
        self._generated_time = generated_time
        self._global_unit_id = global_unit_denomination
        self._custom_unit_id = custom_unit_denomination
        self._tree = tree

        return

    entity = Immutable(lambda s: s._entity)
    session = Immutable(lambda s: s._entity.session)
    balance_time = Immutable(lambda s: s._balance_time.raw)
    generated_time = Immutable(lambda s: s._generated_time.raw)
    custom_unit_id = Immutable(lambda s: s._custom_unit_id)
    global_unit_id = Immutable(lambda s: s._global_unit_id)
    nodes = Immutable(lambda s: s._tree)

    has_accounts = Immutable(lambda s: len(s._tree) > 0)

    income = Immutable(lambda s: s.nodes_of_type(AMType.income))
    expenses = Immutable(lambda s: s.nodes_of_type(AMType.expense))
    assets = Immutable(lambda s: s.nodes_of_type(AMType.asset))
    liabilities = Immutable(lambda s: s.nodes_of_type(AMType.liability))
    equities = Immutable(lambda s: s.nodes_of_type(AMType.equity))

    has_assets = Immutable(lambda s: len(s.nodes_of_type(AMType.asset)) > 0)
    has_liabilities = Immutable(lambda s: len(
        s.nodes_of_type(AMType.liability)) > 0
    )
    has_income = Immutable(lambda s: len(s.nodes_of_type(AMType.income)) > 0)
    has_expenses = Immutable(lambda s: len(s.nodes_of_type(AMType.expense)) > 0)
    has_equity = Immutable(lambda s: len(s.nodes_of_type(AMType.equity)) > 0)

    total_assets = Immutable(lambda s: s._total(AMType.asset))
    total_expenses = Immutable(lambda s: s._total(AMType.expense))
    total_liabilities = Immutable(lambda s: s._total(AMType.liability))
    total_income = Immutable(lambda s: s._total(AMType.income))
    total_equity = Immutable(lambda s: s._total(AMType.equity))

    def nodes_of_type(self, am_type: AMType) -> List[TreeNode]:
        """Return top-level TreeNodes of the supplied AMType"""
        if not isinstance(am_type, AMType):
            raise TypeError('am_type must be of type `AMType`')
        if not self.has_accounts:
            return list()
        nodes = [n for n in self._tree if n.am_type == am_type]
        return nodes

    def _total(self, am_type: AMType) -> Decimal:
        """Return total recursive balance for accounts with supplied type"""
        assert isinstance(am_type, AMType)
        accounts = self.nodes_of_type(am_type)
        if len(accounts) < 1:
            return Decimal(0)
        total = sum([a.recursive_balance for a in accounts])
        assert isinstance(total, Decimal)
        return total

    @classmethod
    def decode(cls: Type[T], entity: Entity, data: Any) -> T:

        if not isinstance(data, dict):
            raise UnexpectedResponseType(data, dict)

        try:

            tree = cls(
                entity=entity,
                balance_time=AmatinoTime.decode(data['balance_time']),
                generated_time=AmatinoTime.decode(data['generated_time']),
                global_unit_denomination=data['global_unit_denomination'],
                custom_unit_denomination=data['custom_unit_denomination'],
                tree=TreeNode.decode_many(entity, data['tree'])
            )

        except KeyError as error:
            raise MissingKey(error.args[0])

        return tree

    @classmethod
    def retrieve(
        cls: Type[T],
        entity: Entity,
        balance_time: datetime,
        denomination: Denomination
    ) -> T:

        arguments = cls.RetrieveArguments(
            balance_time=balance_time,
            denomination=denomination
        )

        return cls._retrieve(entity, arguments)

    @classmethod
    def _retrieve(
        cls: Type[T],
        entity: Entity,
        arguments: K
    ) -> T:

        if not isinstance(entity, Entity):
            raise TypeError('entity must be of type Entity')

        assert isinstance(arguments, cls.RetrieveArguments)
        data = DataPackage(object_data=arguments, override_listing=True)
        parameters = UrlParameters(entity_id=entity.id_)

        request = ApiRequest(
            path=cls._PATH,
            method=HTTPMethod.GET,
            credentials=entity.session,
            data=data,
            url_parameters=parameters
        )

        return cls.decode(entity, request.response_data)

    class RetrieveArguments(Encodable):
        def __init__(
            self,
            balance_time: datetime,
            denomination: Denomination
        ) -> None:

            if not isinstance(balance_time, datetime):
                raise TypeError('balance_time must be of type `datetime`')

            if not isinstance(denomination, Denomination):
                raise TypeError('denomination must be of type `denomination`')

            self._balance_time = AmatinoTime(balance_time)
            self._denomination = denomination

        def serialise(self) -> Dict[str, Any]:

            global_unit_id = None
            custom_unit_id = None

            if isinstance(self._denomination, GlobalUnit):
                global_unit_id = self._denomination.id_
            else:
                assert isinstance(self._denomination, CustomUnit)
                custom_unit_id = self._denomination.id_

            data = {
                'balance_time': self._balance_time.serialise(),
                'custom_unit_denomination': custom_unit_id,
                'global_unit_denomination': global_unit_id
            }

            return data



"""
Amatino API Python Bindings
Transaction Version List Module
Author: hugh@amatino.io
"""
from amatino import Transaction, Entity
from amatino.internal.api_request import ApiRequest
from amatino.internal.url_parameters import UrlParameters
from amatino.internal.url_target import UrlTarget
from amatino.internal.http_method import HTTPMethod
from amatino.unexpected_response_type import UnexpectedResponseType
from amatino.internal.immutable import Immutable
from amatino.api_error import ApiError
from amatino.missing_key import MissingKey
from typing import TypeVar, Type, List, Any
from collections import Sequence

T = TypeVar('T', bound='TransactionVersionList')


class TransactionVersionList(Sequence):
    """
    Amatino retains a version history of every Transaction. That history
    allows you to step backwards and forwards through changes to the accounting
    information describing an Entity. To view the history of a Transaction, you
    can retrieve a Transaction Version List.
    """
    _PATH = '/transactions/version/list'

    def __init__(
        self,
        entity: Entity,
        transaction_id: int,
        versions: List[Transaction]
    ) -> None:

        assert isinstance(entity, Entity)
        assert isinstance(transaction_id, int)
        assert isinstance(versions, list)
        if len(versions) > 0:
            assert False not in [isinstance(t, Transaction) for t in versions]

        self._entity = entity
        self._transaction_id = transaction_id
        self._versions = versions

        return

    versions = Immutable(lambda s: s._versions)
    entity = Immutable(lambda s: s._entity)
    session = Immutable(lambda s: s._entity.session)

    def __len__(self):
        return len(self.versions)

    def __getitem__(self, key):
        return self.versions[key]

    def __iter__(self):
        return TransactionVersionList.Iterator(self._versions)

    class Iterator:
        """An iterator for iterating through versions"""

        def __init__(self, versions: List[Transaction]) -> None:
            self._index = 0
            self._versions = versions
            return

        def __next__(self) -> Transaction:
            if self._index >= len(self._versions):
                raise StopIteration
            version = self._versions[self._index]
            self._index += 1
            return version

    @classmethod
    def retrieve(
        cls: Type[T],
        entity: Entity,
        transaction: Transaction
    ) -> T:
        """Return a TransactionVersionList for the supplied Transaction"""

        if not isinstance(transaction, Transaction):
            raise TypeError('transaction must be of type Transaction')

        if not isinstance(entity, Entity):
            raise TypeError('entity must be of type Entity')

        targets = [UrlTarget.from_integer('transaction_id', transaction.id_)]
        parameters = UrlParameters(entity_id=entity.id_, targets=targets)
        request = ApiRequest(
            path=cls._PATH,
            method=HTTPMethod.GET,
            credentials=entity.session,
            data=None,
            url_parameters=parameters
        )

        return cls._decode(entity, request.response_data)

    @classmethod
    def _decode(cls: Type[T], entity: Entity, data: Any) -> T:

        if not isinstance(data, list):
            raise UnexpectedResponseType(data, list)

        if len(data) < 1:
            raise ApiError('Response unexpectedly empty')

        tx_list_data = data[0]

        try:

            if tx_list_data['versions'] is None:
                tx_list_data['versions'] = list()

            tx_list = cls(
                entity=entity,
                transaction_id=tx_list_data['transaction_id'],
                versions=Transaction.decode_many(
                    entity,
                    tx_list_data['versions']
                )
            )

        except KeyError as error:
            raise MissingKey(error.args[0])

        return tx_list



"""
Amatino API Python Bindings
Account Test Module
Author: hugh@amatino.io
"""
from datetime import datetime
from amatino import Side, Entry, Transaction
from decimal import Decimal
from amatino.tests.primary.account import AccountTest
from amatino import AMType, ResourceNotFound
from typing import Optional

NAME = 'Create, retrieve, update, delete a Transaction'


class TransactionTest(AccountTest):
    """Test the Transaction object"""

    def __init__(self, name=NAME) -> None:

        super().__init__(name)
        self.asset = self.create_account(AMType.asset, 'Test Asset')
        self.liability = self.create_account(AMType.liability, 'Test Liability')
        return

    def create_transaction(
        self,
        time: Optional[datetime] = None,
        amount: Optional[Decimal] = None
    ) -> Transaction:

        if time is None:
            time = datetime.utcnow()
        if amount is None:
            amount = Decimal(10)

        transaction = Transaction.create(
            self.entity,
            datetime.utcnow(),
            [
                Entry(Side.debit, amount, self.asset),
                Entry(Side.credit, amount, self.liability)
            ],
            self.usd,
            'Test transaction'
        )

        return transaction

    def execute(self) -> None:

        amount = Decimal(10)

        try:
            transaction = self.create_transaction(amount=amount)
        except Exception as error:
            self.record_failure(error)
            return

        assert isinstance(transaction, Transaction)

        try:
            transaction = Transaction.retrieve(
                self.entity,
                transaction.id_,
                self.asset.denomination
            )
            assert isinstance(transaction, Transaction)
            assert isinstance(transaction.time, datetime)
            assert isinstance(transaction.version_time, datetime)
            assert isinstance(transaction.entries, list)
            assert transaction.magnitude == Decimal(10)
            assert len(transaction) > 1
            for entry in transaction:
                assert isinstance(entry, Entry)

        except Exception as error:
            self.record_failure(error)
            return

        updated_description = 'Updated test description goodness'

        try:
            transaction = transaction.update(description=updated_description)
        except Exception as error:
            self.record_failure(error)
            return

        if not transaction.description == updated_description:
            self.record_failure('Transaction description not updated')
            return

        try:
            transaction.delete()
        except Exception as error:
            self.record_failure(error)
            return

        try:
            Transaction.retrieve(
                self.entity,
                transaction.id_,
                self.asset.denomination
            )
        except Exception as error:
            if not isinstance(error, ResourceNotFound):
                self.record_failure(error)
                return

        self.record_success()
        return
